package SAS::Header;

=head1 NAME

SAS::Header - Create a documentation header section for a SAS program

=head1 SYNOPSIS

 use SAS::Header;
 $p = new SAS::Header;
 $p->parse_file('mysas.sas');       # returns a SAS::Parser object
 $header = $p->makeheader();        # extract info and format the header

= head1 DESCRIPTION

I<SAS::Header> is a sub-class of SAS::Parser which parses a SAS file
and creates a block-comment header.

=head2 Methods

The following methods are defined in the SAS:Header class.

=over 4

=item $p -> parse_file($filename)

Calls the SAS::Parser parse_file() method with the options
silent=>1, trim=>0, store=>'global|ccomment'.  

=item $header = $p -> makeheader()

Call makeheader() after the file has been parsed.  makeheader()
extracts the following information from the SAS::Header object:
name, title, procs called, datasets created, macros defined, macros
called, author, created, revised, version.

Each item is formatted as a line of the form:

    Key: Information

e.g.,

   Name: filename.sas
  Title: Description of what I do

and then the collection of lines is formatted as a boxed /* ... */
comment, whose width is determined by the variable C<$SAS::Header::width>.
Any Information portion which would exceed C<$SAS::Header::width>
is wrapped so that successive lines are
indented.

This boxed comment is returned as a "\n" separated string.

=item $p->parse_ccomment($statement)

This method is used to override the default parse_ccomment() method
of the SAS::Parser class, so that any information present in an
existing program header may be found during the initial parsing.

As defined here, this method ignores all C-comments contained within a
PROC or DATA step.  For other C-comments, it looks for strings of the
following types:

  Author:
  Title:
  Created:
  Version:

The keyword may be in upper, lower, or mixed case, but must be followed
by (optional whitespace) and a ':'.  When such a line is found, a
corresponding entry is added to the SAS::Header object, and later used
by makeheader().  E.g., the Author information would be stored or
accessed as C<$p-E<gt>{author}>.

=item $author = $p->get_author()

Returns an author string extracted from the parse.  If an author was found
by parse_ccomment(), that string is returned.  Otherwise, the get_author()
method assumes the current user is the author, and returns a string
composed from the USER and HOST environment variables.


=item $title = $p->get_title()

Returns a title string extracted from the parse.  If a title was found
by parse_ccomment(), that string is returned.  Otherwise, the get_title()
method tries harder, by examining the $p->stored() statements, 
-- either /* ... */ comments, or TITLE statements
until something reasonable is found.

=item $version = $p->get_version()

Returns a version string extracted from the parse.  If a version was found
by parse_ccomment(), that string is returned.  Otherwise, the get_version()
method returns undef.

=back 4

=head2 Formatting C-comments

The following subroutines are used to process /* ... */ comments:

=over 4

=item $boxed = &box($text)

Creates a boxed multi-line /* ... */ comment where each line is of width
C<$SAS::Header::width>.

=item $unboxed = &unbox($text)

Takes a string containing a boxed multi-line /* ... */ comment and removes
the decorator characters and an initial and trailing line of decorators.

=back 4

=cut


use SAS::Parser;
@ISA = qw(SAS::Parser);

use vars qw($VERSION $DEBUG $width);

$VERSION = sprintf("%d.%02d", q$Revision: 0.8 $ =~ /(\d+)\.(\d+)/);
BEGIN	{
	$width = 70;  # <= header width
	$DEBUG = 0;
}

sub new
{
    my($class) = @_;

    my $self = $class->SUPER::new;
    $self;
}

sub parse_file {
	my $self = shift;
	my $file = shift;
	my %options = (
		silent => 1,
		trim   => 0,
		store  => 'global|ccomment',
		);
	$self->SUPER::parse_file($file, \%options);
	$self;
}

## Override the parse_comment method to extract information from an
## existing header: Author, Title, Created, Version.

#  For efficiency, we stop trying once a DATA or PROC step has been seen.

sub parse_ccomment {		# $self->parse_ccomment($statement);
   my($self, $stmt) = @_;
	
	return unless $step eq '';

	$stmt = &unbox($stmt);
	my @lines = split(/\n/, $stmt);
	foreach (@lines) {
		if (/(author|title|created|version)\s*:\s*/i) {
			my $info = lc($1);
			my $rest = $';
			$rest =~ s/\s*$//;
			$rest =~ s/\s+/ /g;
			#print "$info :: $rest\n";
			$self->{$info} = $rest;
		};
	}
}


sub makeheader {
	my $self = shift;
	my $head;

	# standard information
	my $file = $self->{file};
	my @procs = $self->procs();
	my @macros = $self->macros();
	my @macdefs = $self->macdefs();
	
	# additional info, for SAS::Header
	my $date = (stat $file)[9];
	my $title = $self->get_title();
	my $author = $self->get_author();
	my $version = $self->get_version();
		
	my $sep = '-' x $width . "\n";
	$head  = &fmt('name', $file)
			 . &fmt('title', $title)
			 . $sep;
	$head .= &fmt('procs', join(' ', @procs))	      if scalar @procs;
	$head .= &fmt('datasets', join(' ', @datasets)) if scalar @datasets;
	$head .= &fmt('macdefs', join(' ', @macdefs))   if scalar @macdefs;
	$head .= &fmt('macros', join(' ', @macros))     if scalar @macros;
	$head .= $sep;
	$head .= &fmt('author', $author) if $author;
	
	$head .= (defined($self->{created})) ?
				&fmt('created', $self->{created}) 
			 : &fmt('created', date($date));
	$head .= &fmt('revised', date($date))			
			;
	$head .= &fmt('version', $version) if $version;

	return &box($head, $width, 1);
}

sub fmt {
	my ($tag, $text) = @_;

	my $t = sprintf("%8s: %s\n", ucfirst($tag), $text);
	local $columns = $width - 5;
	$t = &wrap("", " " x 10, $t);
	return $t;
}

# We use Text::Wrap to fold long lines, but suppress the transformation
# of tabs back to spaces by re-defining unexpand as a no-op.

#use Text::Wrap;

use Text::Tabs qw(expand);
sub unexpand {return shift;};

## from Text::Wrap, but without tab re-expansion
sub wrap
{
	my ($ip, $xp, @t) = @_;

	my $r = "";
	my $t = expand(join(" ",@t));
	my $lead = $ip;
	my $ll = $columns - length(expand($lead)) - 1;
	my $nl = "";

	# remove up to a line length of things that aren't
	# new lines and tabs.

	if ($t =~ s/^([^\n]{0,$ll})(\s|\Z(?!\n))//xm) {

		# accept it.
		$r .= unexpand($lead . $1);

		# recompute the leader
		$lead = $xp;
		$ll = $columns - length(expand($lead)) - 1;
		$nl = $2;

		# repeat the above until there's none left
		while ($t) {
			if ( $t =~ s/^([^\n]{0,$ll})(\s|\Z(?!\n))//xm ) {
				print "\$2 is '$2'\n" if $debug;
				$nl = $2;
				$r .= unexpand("\n" . $lead . $1);
			} elsif ($t =~ s/^([^\n]{$ll})//) {
				$nl = "\n";
				$r .= unexpand("\n" . $lead . $1);
			}
		}
		$r .= $nl;
	} 

	die "couldn't wrap '$t'" 
		if length($t) > $ll;

	print "-----------$r---------\n" if $debug;

	print "Finish up with '$lead', '$t'\n" if $debug;

	$r .= $lead . $t if $t ne "";

	print "-----------$r---------\n" if $debug;
	return $r;
}



# Create a C-style boxed comment. Assumes all lines have been previously
# folded if necessary to fit given width.

sub box {
	my ($text, $width, $in) = @_;
	my @lines = split ("\n", $text);
	my ($line, $boxed);
	
	my $boxed;
	$boxed  = ' ' x $in . '/*' . '-' x ($width-$in-4) . "*\n";
	foreach $line (@lines) {
		$line = ' ' . $line . ' ' x $width;
		$line = substr($line,0, ($width-$in-4));
		$boxed .= ' ' x $in . ' *' . $line             . "*\n";
	}
	$boxed .= ' ' x $in . ' *' . '-' x ($width-$in-4) . "*/\n";
	return $boxed;
}

# format a filemod time, as 17-Jan-97 12:34
sub date {

    local($time) = shift || time();
    local($[) = 0;
    local($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst);
    local(@MoY) = ('Jan','Feb','Mar','Apr','May','Jun',
	    'Jul','Aug','Sep','Oct','Nov','Dec');
    ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($time);
	 $year = $year+1900;     # Y2K !!
    sprintf("%02d %3s %4d %02d:%02d:%02d", 
	 	$mday, $MoY[$mon], $year, $hour, $min, $sec);
}

# Find something reasonable to use for a title from the stored statements
# -- either in a leading /* ... */ comment, or in a title statement,
# unless a title has already been defined.
# First one wins.  Assumes that global and ccomment statements were
# stored.

sub get_title {
	my $self = shift;
	my $title;

	return $self->{title} if defined($self->{title});	

	my @stored = $self->stored();
	foreach $s (@stored) {
		my ($lineno, $step, $type, $stmt, $statement) = @$s;

		# find first line in a C-comment containing 'Title:'
		if ($type eq 'ccomment' && $statement =~ /title\s*:/i) {
			$title = $';
			($title) = (split(/\n/, $title))[0];
			$title =~ s/^\s*//;
			$title =~ s{\s*[|*/]?\s*$}{};
			last;
		}
		# or, the first title statement
		elsif ($stmt eq 'title') {
			($title = $statement) =~ s/title\d?\s+['"]?//;
			$title =~ s/['"]?\s*;.*$//;
			last;
		}
	}
#	chomp($title);
	return $title;
}

sub get_author {
	my $self = shift;
	# if we have already seen an Author:, return that.
	return $self->{author}  if defined($self->{author});
	
	# otherwise, get information from environment, assuming current user
	# is the author.
	my $user = $ENV{USER};
	my $host = $ENV{HOST} || `hostname`;
	my $author ="<$user\@$host>";

	# Try to use nidump to get name from netinfo database
	# On AIX there's the lsuser command, but with different format
	if (-x '/usr/bin/nidump') {
		my $ninfo = `nidump passwd .`;
		#print "Trying nidump for $author\n";
		foreach (split(/\n/, $ninfo)) {
			($u, $name) = (split(':'))[0, 4];
			if ($u eq $user) {
				$author = $name . " $author";
				last;
			}
		}
	}
	return $author;	
}

sub get_version {
	my $self = shift;

	return $self->{version}  if defined($self->{version});
	return undef;
}	

# Remove decorators around a boxed comment

sub unbox {
	my $text = shift;
	my @lines = split ("\n", $text);
	
	unshift @lines if $lines[0] =~ m{(#|/\*)[\s*-=#]{2,}$};
	pop @lines if $lines[-1] =~ m{[\s*-=#]{2,}\*/};
	
	foreach (@lines) {
		s|^\s*/*\s*||;
		s|\s*\*/\s*$||;
		s/^\s*[*-=|#]+\s*//;
		s/\s*[*-=|#]+\s*$//
	}
	return join("\n",@lines) . "\n";
}

1;
